Package: RITCH
Type: Package
Title: Parser for the ITCH Protocol
Version: 0.1.30
Authors@R: c(
    person("David", "Zimmermann-Kollenda", , "david_j_zimmermann@hotmail.com", role = c("aut", "cre"))
  )
Description: Efficiently parses, filters, and writes binary ITCH files (Version 5.0) containing detailed financial transactions as distributed by NASDAQ to a data.table. Includes functions to interact with NASDAQ data services at <https://emi.nasdaq.com/ITCH/> and <https://emi.nasdaq.com/ITCH/Stock_Locate_Codes/>.
License: MIT + file LICENSE
URL: https://davzim.github.io/RITCH/, https://github.com/DavZim/RITCH
BugReports: https://github.com/DavZim/RITCH/issues
Depends: R (>= 3.5.0)
Imports: data.table, Rcpp (>= 0.12.12), nanotime (>= 0.3.2), bit64 (>=
        4.0.5)
LinkingTo: Rcpp
Encoding: UTF-8
RoxygenNote: 7.3.3
Suggests: tinytest
NeedsCompilation: yes
Packaged: 2026-04-16 14:49:26 UTC; dazimme
Author: David Zimmermann-Kollenda [aut, cre]
Maintainer: David Zimmermann-Kollenda <david_j_zimmermann@hotmail.com>
Repository: CRAN
Date/Publication: 2026-04-21 18:52:28 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-26 01:45:53 UTC; unix
Archs: RITCH.so.dSYM
