# Generated by roxygen2: do not edit by hand

export(add_meta_to_filename)
export(count_ipo)
export(count_luld)
export(count_market_participant_states)
export(count_messages)
export(count_modifications)
export(count_mwcb)
export(count_noii)
export(count_orders)
export(count_reg_sho)
export(count_rpii)
export(count_stock_directory)
export(count_system_events)
export(count_trades)
export(count_trading_status)
export(download_sample_file)
export(download_stock_directory)
export(filter_itch)
export(format_bytes)
export(get_date_from_filename)
export(get_exchange_from_filename)
export(get_modifications)
export(get_msg_classes)
export(get_orders)
export(get_trades)
export(gunzip_file)
export(gzip_file)
export(list_sample_files)
export(open_itch_sample_server)
export(open_itch_specification)
export(read_ipo)
export(read_itch)
export(read_luld)
export(read_market_participant_states)
export(read_modifications)
export(read_mwcb)
export(read_noii)
export(read_orders)
export(read_reg_sho)
export(read_rpii)
export(read_stock_directory)
export(read_system_events)
export(read_trades)
export(read_trading_status)
export(write_itch)
import(data.table)
importFrom(Rcpp,sourceCpp)
importFrom(bit64,as.integer64)
importFrom(nanotime,nanotime)
importFrom(utils,browseURL)
importFrom(utils,download.file)
useDynLib(RITCH)
